from dawgs_ml.dataframe import dataframe as df
import utils


class Split:
    def __init__(self, data: df.DataFrame, target: str | int, seed=1):
        """Class to split data into train and test sets. If random seed parameter is not provided or is set to some
        non-zero value, then the dataframe is shuffled before splitting. If the random seed is set to 0, then no
        random shuffling is performed and the dataframe is sliced at the nth position(s) to create the other subsets."""
        self.data = data.copy()
        self.target = target
        self.random_seed = seed
        self.X_train = None
        self.X_test = None
        self.Y_train = None
        self.Y_test = None
        self.X_val = None
        self.Y_val = None
        if self.random_seed != 0:
            self.data.shuffle(seed=self.random_seed)

    def train_test(self, test_size: float = 0.2) -> (df.DataFrame, df.Column, df.DataFrame, df.Column):
        """Method to split data into train and test sets.

        Args:
            test_size (float): proportion of the dataset to include in the test split. Default is 0.2

        Returns:
            values (tuple): 4 values in the following order:
            X_test, Y_test, X_train, Y_train
            """
        total_rows = self.data.rows
        test_size = int(total_rows * test_size)
        test_set = df.DataFrame()  # initialize empty test set dataframe
        test_set.set_columns_name(self.data.get_column_names())

        for i in range(test_size):
            test_set.add_row(self.data.pop_row(i))

        # After popping,the remaining rows in self.data is the train set
        self.Y_test = test_set.pop_column(self.target)
        self.X_test = test_set
        self.Y_train = self.data.pop_column(self.target)
        self.X_train = self.data

        return self.X_test, self.Y_test, self.X_train, self.Y_train

    def train_test_val(self, test_size=0.2, val_size=0.1):
        """Method to split data into train, test, and validation sets.

        Args:
            test_size (float): proportion of the dataset to include in the test split. Default is 0.2
            val_size (float): proportion of the dataset to include in the validation split. Default is 0.1

        Returns:
            values (tuple): 6 values in the following order:
            X_test, Y_test, X_train, Y_train, X_val, Y_val
            """
        total_rows = self.data.rows
        test_size = int(total_rows * test_size)
        val_size = int(total_rows * val_size)
        test_set = df.DataFrame()  # initialize empty test set dataframe
        val_set = df.DataFrame()  # initialize empty validation set dataframe
        test_set.set_columns_name(self.data.get_column_names())
        val_set.set_columns_name(self.data.get_column_names())

        for i in range(test_size):
            test_set.add_row(self.data.pop_row(i))

        for i in range(val_size):
            val_set.add_row(self.data.pop_row(i))

        # After popping, the remaining rows in self.data is the train set
        self.Y_test = test_set.pop_column(self.target)
        self.X_test = test_set
        self.Y_val = val_set.pop_column(self.target)
        self.X_val = val_set
        self.Y_train = self.data.pop_column(self.target)
        self.X_train = self.data
        return self.X_test, self.Y_test, self.X_train, self.Y_train, self.X_val, self.Y_val


class KFoldCrossValidation:
    def __init__(self, data: df.DataFrame, target_column: str | int, k: int, seed: int = 1):
        self._k = k
        self._data = data
        self._target_column = target_column
        self._s = seed
        self._folds = []
        if k <= 1 or k >= self._data.rows:
            raise ValueError("k must be greater than 1 and less than the number of rows in the dataframe")

    @property
    def folds(self) -> list[tuple[df.DataFrame, df.Column]]:
        return self._folds

    def split(self):
        """Method to split the dataframe into k folds. It creates an instance of the Split class, and it iteratively
        calls the train_test method to split the dataframe into k folds, by adjusting the ratio at every iteration so
        to make folds of the same size. The leftover is the last fold so that train_test split is called k-1 times.
        It returns a list of folds [(X1_test, Y1_test), (X2_test, Y2_test), ...]. """
        ratio = 1 / self._k
        split_size = int(self._data.rows / self._k)
        resizable_data = self._data
        for i in range(self._k - 1):
            split = Split(data=resizable_data, target=self._target_column, seed=self._s)
            X_test, Y_test, X_train, Y_train = split.train_test(test_size=ratio)
            self._folds.append((X_test, Y_test))
            resizable_data = X_train
            resizable_data.add_column(Y_train)
            ratio = split_size / self._data.rows  # update the ratio for split for next iteration
        last_fold_Y = resizable_data.pop_column(self._target_column)
        self._folds.append((resizable_data, last_fold_Y))


class Evaluation:
    def __init__(self, predictions, actual):
        self.predictions = predictions
        self.actual = actual
        self.tp, self.fp, self.fn, self.tn = self._tp_fp_fn_tn()

    def _tp_fp_fn_tn(self):
        tp, fp, fn, tn = (0,) * 4
        for i in range(len(self.predictions)):
            if self.predictions[i] == 1 and self.actual[i] == 1:
                tp += 1
            elif self.predictions[i] == 1 and self.actual[i] == 0:
                fp += 1
            elif self.predictions[i] == 0 and self.actual[i] == 1:
                fn += 1
            else:
                tn += 1
        return tp, fp, fn, tn

    def accuracy(self):
        """Returns the accuracy of the model on the test set, this metric for model evaluation is recommended for
        balanced classes."""
        correct = 0
        for i in range(len(self.predictions)):
            if self.predictions[i] == self.actual[i]:
                correct += 1
        return correct / len(self.predictions)

    def precision(self):
        """Returns the precision of the model on the test set, this metric for model evaluation is recommended for both
        imbalanced and balanced classes."""
        return self.tp / (self.tp + self.fp)

    def recall(self):
        """Returns the recall of the model on the test set, this metric for model evaluation is recommended for both
        imbalanced and balanced classes."""
        return self.tp / (self.tp + self.fn)

    def f1_score(self):
        """Returns the harmonic mean of the precision and recall of the model on the test set, this metric for model
        evaluation is recommended for both imbalanced and balanced classes."""
        return (2 * self.precision() * self.recall()) / (self.precision() + self.recall())

    def confusion_matrix(self) -> list[list[int]]:
        """Returns:
            list[list[int]]: confusion matrix of the model on the test set."""
        print("\t\t\tPredicted - | Predicted + |")
        print(f"Actual - |\t\t {self.tn} \t\t {self.fp}")
        print(f"Actual + |\t\t {self.fn} \t\t\t {self.tp}")
        return [[self.tn, self.fp], [self.fn, self.tp]]


class Combination:
    """Class to store the metrics of each combination."""

    def __init__(self, target_metric: str = "accuracy"):
        self.target_metric = target_metric
        self._fold = _Folds()
        self.combination = {}
        self.best_metric_value = float('-inf')

    def _define_best(self, folds: '_Folds', combination: dict):
        if self.target_metric not in folds.metrics:
            raise ValueError(f"{self.target_metric} is not a recognized metric.")

        # Calculate the new average metric value.
        if folds.metrics[self.target_metric]:
            total = sum(folds.metrics[self.target_metric])
            count = len(folds.metrics[self.target_metric])
            new_metric_value = total / count
        else:
            new_metric_value = float('-inf')  # default value if no metrics exists

        if new_metric_value > self.best_metric_value:   # update the best combination-metric if the new metric is better
            self.best_metric_value = new_metric_value
            self._fold = folds
            self.combination = combination

    def _get_best_combination(self):
        """Returns the best combination and its average metric value."""
        if not self.combination:
            return None, None  # Handle the case where no combination has been defined yet.
        return self.combination, self.best_metric_value

    def __repr__(self):
        return str(f"Best combination: {self.combination}, Average {self.target_metric}: {self.best_metric_value}")


class _Folds:
    """Helper class to store the metrics of each fold."""
    def __init__(self):
        self.metrics = {
            'accuracy': [],
            'precision': [],
            'recall': [],
            'f1_score': []
        }

    def update(self, accuracy, precision, recall, f1_score):
        self.metrics['accuracy'].append(accuracy)
        self.metrics['precision'].append(precision)
        self.metrics['recall'].append(recall)
        self.metrics['f1_score'].append(f1_score)

    def get_average(self, metric):
        if metric not in self.metrics:
            raise ValueError(f"Metric {metric} is not valid.")
        if not self.metrics[metric]:
            return None  # or raise an error if preferred
        return sum(self.metrics[metric]) / len(self.metrics[metric])

    def __repr__(self):
        return str(self.metrics)


class GridSearch:
    """Given a dataframe and a list of hyperparameters, this class will return the best possible combination of
    hyperparameters by running a grid search with k-fold cross validation."""

    def __init__(self, data: df.DataFrame, model_instance, k_folds: int = 5, **hyperparameters):
        self.data = data.copy()  # use the deep copy of the dataframe to avoid changing the original dataframe
        self.model_instance = model_instance
        self.k_folds = k_folds
        if self._correctness_model_params(hyperparameters):
            self.hyperparameters = hyperparameters
        else:
            raise AttributeError("Incorrect hyperparameters for the model instance.")

    def _correctness_model_params(self, hyperparameters):
        """Helper method to check if the hyperparameters of the model instance are correct."""
        for key, value in hyperparameters.items():
            if not hasattr(self.model_instance, key):
                return False
        return True

    def _set_model_params(self, parameters_combination: list):
        """Helper method to set the hyperparameters of the model instance."""
        for i in (range(len(parameters_combination))):
            setattr(self.model_instance, list(self.hyperparameters)[i], parameters_combination[i])

    def _create_train_val_fold(self, target_col):
        """Helper method to create train and validation folds for k-fold cross validation."""
        K = KFoldCrossValidation(data=self.data, target_column=target_col, k=self.k_folds)
        K.split()
        sets = []
        for fold in K.folds:
            temp_df, temp_target = df.DataFrame(), df.Column(target_col)
            validation = fold  # leave one fold out for evaluation on validation set
            for leftover_fold in K.folds:
                if leftover_fold != fold:
                    temp_df.merge(leftover_fold[0])
                    temp_target.name = leftover_fold[1].name
                    temp_target.append_column(leftover_fold[1])
            sets.append((temp_df, temp_target, validation[0], validation[1]))
        return sets

    def return_model_instance(self):
        """Return the model instance as a dictionary with the parameters and their values."""
        return self.model_instance.__dict__

    def run_training(self, target_col: str | int, metric: str = "accuracy"):
        """This method runs the grid search with k-fold cross validation, and it returns the best combination of
        hyperparameters based on the specified metric.

        Args:
            target_col: The name of the target column in the dataframe.
            metric: The metric to use for evaluating the best combination. Default is accuracy.
            """
        folds = self._create_train_val_fold(target_col)
        hyperparameters_combinations = utils.cartesian_product(*self.hyperparameters.values())
        best_combination = Combination(target_metric=metric)
        for combination in hyperparameters_combinations:
            print(f"Running K-fold cross validation with hyperparameters: {combination}")
            fold_metric = _Folds()
            for fold in folds:
                self._set_model_params(combination)
                print(f"Fold: {folds.index(fold) + 1}")
                self.model_instance.fit(fold[0], fold[1], fold[0], fold[1])
                validation_predictions = self.model_instance.predict(fold[2])
                metrics = Evaluation(validation_predictions, fold[3])
                fold_metric.update(metrics.accuracy(), metrics.precision(), metrics.recall(), metrics.f1_score())
                print(f"Accuracy updated to: {fold_metric.get_average(metric)}")
                self.model_instance.__init__()  # reset model parameters after each combination
            best_combination._define_best(fold_metric, combination)
        print(best_combination)
        return best_combination



class RandomSearch:
    """Given a range of continuous values hyperparameters, this class will random sample hyperparameters using the
    k-fold cross validation."""
    pass


class BayesianSearch:
    """Use bayesian optimization to optimize hyperparameters, by focusing more on the most promising hyperparameters"""
    pass



from dataframe import dataframe as df  # this is my custom data structure for dataframes
from utils import utils
from utils.model_utils import Evaluation
import random
from abc import ABC, abstractmethod


class Model(ABC):
    def __init__(self, random_seed: int = 1):
        self._X = None  # the actual data (features only), expecting an instance of DataFrame
        self._Y = None  # the target column, an instance of Column, or a python list
        self._X_val = None
        self._Y_val = None
        self.random_seed = random_seed

    @abstractmethod
    def _predict(self, x: list):
        pass

    @abstractmethod
    def _predictions(self):
        pass

    @abstractmethod
    def _validation_predictions(self):
        pass

    @abstractmethod
    def fit(self, train_data: df.DataFrame, train_target: df.Column | list, test_data: df.DataFrame,
            test_target: df.Column | list):
        pass

    @abstractmethod
    def predict(self, validation_x: df.DataFrame) -> list:
        pass


class Perceptron(Model):
    def __init__(self, learning_rate=0.1, epochs=10, activation: float = 0, verbosity: bool = False):
        super().__init__()
        self._weights = []
        self._step_activation = activation
        self.verbosity = verbosity
        self._bias = 0  # introduced bias for non-linearly separable problems having to pass through the origin at init
        self.learning_rate = learning_rate
        self.epochs = epochs
        self._errors = []
        self._accuracy = 0
        self.train_accuracy = 0
        self.test_accuracy = 0
        random.seed(self.random_seed)

    @property
    def weights(self) -> list:
        return self._weights

    @weights.setter
    def weights(self, weights: list):
        self._weights = weights

    @property
    def X(self) -> df.DataFrame:
        return self._X

    @property
    def Y(self) -> list:
        return self._Y

    @property
    def X_val(self) -> df.DataFrame:
        return self._X_val

    @property
    def Y_val(self) -> list:
        return self._Y_val

    @property
    def step_activation(self):
        return self._step_activation

    @step_activation.setter
    def step_activation(self, step_activation):
        self._step_activation = step_activation

    def activation(self, value):
        return 1 if value > self._step_activation else 0

    def initialize_weights(self):
        for _ in range(self._X.columns):  # for each self.data.columns:
            weight = (random.random() - 0.5) * 0.1  # random value between -0.05 and 0.05
            self._weights.append(weight)

    def fit(self, train_data: df.DataFrame, train_target: df.Column | list, test_data: df.DataFrame,
            test_target: df.Column | list):
        """ Fit the model to the training data, accuracy is calculated on the validation data, but if you do not have
        a validation data, then use the training data again as validation data, especially in the context of just having
        the need to train a model on the whole dataset after running a Search algorithm for hyperparameters in k-fold.

        Args:
            train_data: The training data.
            train_target: The training target.
            test_data: The test data.
            test_target: The test target.
        """
        self._X = train_data
        self._Y = train_target
        self._X_val = test_data
        self._Y_val = test_target
        self._bias = 0
        self._weights = []
        self.initialize_weights()

        # fit the model
        for epoch in range(self.epochs):
            error = 0
            for i in range(self._X.rows):
                x_row = self._X[i]
                y = self._Y[i]
                y_prediction = self._predict(x_row)
                if y != y_prediction:
                    error += 1
                    for j in range(len(x_row)):  # update the weights
                        self._weights[j] += self.learning_rate * (y - y_prediction) * x_row[j]  # update rule
                    self._bias += self.learning_rate * (y - y_prediction)  # update the bias
            train_prediction = self._predictions()
            val_prediction = self._validation_predictions()
            train_accuracy = Evaluation(train_prediction, self._Y).accuracy()
            val_accuracy = Evaluation(val_prediction, self._Y_val).accuracy()
            if self.verbosity:
                print(f"Epoch: {epoch} || Error: {error}, Bias: {round(self._bias, 4)}, "
                      f"Train accuracy: {round(train_accuracy, 4)}, Validation accuracy: {round(val_accuracy, 4)}")

    def _predict(self, x):
        """Helper function that predicts the output for a given input and returns 0 or 1 based on activation step"""
        y = 0
        for i in range(len(x)):
            y += self._weights[i] * x[i]
        y += self._bias
        return self.activation(y)

    def _predictions(self):
        """Helper function that uses the trained model, after fitting, to make predictions on the train data."""
        predictions = []
        for i in range(self._X.rows):
            predictions.append(self._predict(self._X[i]))
        return predictions

    def _validation_predictions(self):
        """Helper function that uses the trained model, after fitting, to make predictions on the validation data."""
        predictions = []
        for i in range(self._X_val.rows):
            predictions.append(self._predict(self._X_val[i]))
        return predictions

    def predict(self, validation_x: df.DataFrame) -> list:
        """It uses the trained model, after fitting, to make predictions on the validation data."""
        predictions = []
        for i in range(validation_x.rows):
            value = utils.dot_product(validation_x[i], self._weights)
            value += self._bias
            predictions.append(self.activation(value))
        return predictions



from dataframe import dataframe as df


class OneHotEncoder:
    def __init__(self):
        pass

    def fit(self, X: df.DataFrame, Y: df.Column | list):
        pass

    def transform(self, X: df.DataFrame) -> df.DataFrame:
        pass


class MinMaxScaler:
    def __init__(self):
        pass

    def fit(self, X: df.DataFrame, Y: df.Column | list):
        pass

    def transform(self, X: df.DataFrame) -> df.DataFrame:
        pass


class StandardScaler:
    def __init__(self):
        pass

    def fit(self, X: df.DataFrame, Y: df.Column | list):
        pass

    def transform(self, X: df.DataFrame) -> df.DataFrame:
        pass


from dawgs_ml.dataframe import dataframe as df


def dot_product(v1, v2):
    return sum(x * y for x, y in zip(v1, v2))


def step_activation(x):
    if x > 0:
        return 1
    else:
        return 0


def cartesian_product(*arrays):
    if len(arrays) == 1:
        return [[item] for item in arrays[0]]  # base case, single list
    else:  # recursive case
        result = []
        remaining = cartesian_product(*arrays[1:])
        for item in arrays[0]:
            for r in remaining:
                result.append([item] + r)
        return result



from setuptools import setup, find_packages

setup(
    name='DawgsML',
    version='0.0.2',
    author='Andrea Marchi',
    author_email='andreaphilly@hotmail.com',
    description='A simple library for machine learning without a requirements.txt',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Doriasamp/DawgsML.git',
    packages=find_packages(),
    install_requires=[],  # nothing here
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    python_requires='>=3.9',
    keywords='machine learning, data processing, custom data structures, pandas-like, data analysis, ml, '
             'data science, statistics, numpy, deep learning, preprocessing, pandas simulation',
    project_urls={
        'Source': 'https://github.com/Doriasamp/DawgsML.git',
    },
)



"""A library for creating and manipulating dataframes in a tabular format.
Basically a kinkier and a (much) slower version of pandas used for all the data science projects so that I can keep my
promise of building a machine leaning library without a requirements.txt file."""
import random


class Column:
    def __init__(self, name: str):
        self._name = name
        self._data = []
        self._type = "str"

    @property
    def data(self) -> list:
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, data_type: str):
        if self._type == data_type:
            return
        match data_type:
            case "int":
                for i in range(len(self._data)):
                    self._data[i] = int(self._data[i])
            case "float":
                for i in range(len(self._data)):
                    self._data[i] = float(self._data[i])
            case "str":
                pass
            case _:
                raise ValueError("Invalid data type")
        self._type = data_type

    def add_data(self, data):
        self._data.append(data)

    def count_occurrences(self, value_to_count):
        count = 0
        for value in self._data:
            if value == value_to_count:
                count += 1
        return count

    def append_column(self, other: 'Column'):
        for i in range(len(other)):
            self._data.append(other.data[i])

    def dot_product(self, other) -> float | int:
        if len(self) != len(other):
            raise ValueError("Columns must have the same length")

        if isinstance(other, Column):
            if not all(isinstance(x, (int, float, bool)) for x in self.data) or not all(
                    isinstance(y, (int, float, bool)) for y in other.data):
                raise ValueError("Columns must be entirely numeric")
            return sum(x * y for x, y in zip(self.data, other.data))

        if isinstance(other, list):
            if not all(isinstance(x, (int, float, bool)) for x in self.data) or not all(
                    isinstance(y, (int, float, bool)) for y in other):
                raise ValueError("Columns must be entirely numeric")
            return sum(x * y for x, y in zip(self.data, other))

    def __len__(self):
        return len(self._data)

    def __mul__(self, other):
        """Overloaded * operator for multiplying a column by a scalar or another column.
        If other is an instance of Column, the dot product is calculated and returned"""
        if isinstance(other, int) or isinstance(other, float):
            for i in range(len(self._data)):
                self._data[i] *= other
        elif isinstance(other, Column) or isinstance(other, list):
            return self.dot_product(other)

    def __rmul__(self, other):
        """Right multiplication operator"""
        return self.__mul__(other)

    def __str__(self):
        return f"<Column {self.name}; {self.type}; {len(self)}> | {self._data[:10]}"

    def __repr__(self):
        return f"<Column '{self._name}' with {len(self._data)} rows>"

    def __getitem__(self, item: int):
        return self._data[item]

    def __setitem__(self, key: int, value):
        self._data[key] = value

    def __iter__(self):
        return iter(self._data)


class Row:
    """Proxy class only used internally by the DataFrame class"""

    def __init__(self, dataframe, row_index):
        self._dataframe = dataframe
        self._row_index = row_index

    def __getitem__(self, column_index):
        return self._dataframe._data[column_index].data[self._row_index]

    def __setitem__(self, column_index, value):
        self._dataframe._data[column_index].data[self._row_index] = value

    def __len__(self):
        return self._dataframe._columns

    def __iter__(self):
        return (self._dataframe._data[col].data[self._row_index] for col in range(self._dataframe._columns))

    def __repr__(self):
        return str([self.__getitem__(col) for col in range(self._dataframe._columns)])


class DataFrame:
    def __init__(self, source=None):
        if source is None:
            source = {}
        self._input_data = source
        self._data = []  # list of Column objects
        self._rows = 0  # read only
        self._columns = 0  # read only
        if source:
            self.__extract_data_from_dict()

    @property
    def rows(self) -> int:
        return self._rows

    @property
    def columns(self) -> int:
        return self._columns

    @rows.setter
    def rows(self, rows: int):
        raise AttributeError("rows is read only")

    @columns.setter
    def columns(self, columns: int):
        raise AttributeError("columns is read only")

    def column(self, name: str) -> Column:
        for column in self._data:
            if column.name == name:
                return column
        raise ValueError("Column not found")

    def __extract_data_from_dict(self):
        if len(self._input_data) == 0:
            return  # empty dictionary
        for key, value in self._input_data.items():
            column_obj = Column(key)
            column_obj.data = value
            self._data.append(column_obj)
        self._rows = len(self._data[0].data)
        self._columns = len(self._data)

    def __repr__(self):
        return self.__str__()

    def type(self, data_type: str):
        match data_type:
            case "int":
                for i in range(len(self._data)):
                    self._data[i].type = data_type
            case "float":
                for column in self._data:
                    column.type = data_type
            case "str":
                for column in self._data:
                    column.type = data_type
            case _:
                raise ValueError("Invalid data type")

    def shape(self) -> tuple:
        """Returns a tuple of (rows, columns)"""
        return self._rows, self._columns

    def copy(self):
        """Create a deep copy of the dataframe"""
        new_df = DataFrame()
        new_df.set_columns_name(self.get_column_names())
        new_df.merge(self)
        return new_df

    def head(self, n: int = 5):
        string = ""
        for column in self._data:
            string += f"{column.name}\t"
        string += "\n"
        for i in range(n):
            for column in self._data:
                string += f"{column.data[i]}\t"
            string += "\n"
        print(string)

    def print(self):
        string = ""
        for column in self._data:
            string += f"{column.name}\t"
        string += "\n"
        for i in range(len(self._data[0])):
            for column in self._data:
                string += f"{column.data[i]}\t"
            string += "\n"
        print(string)

    def add_row(self, row: list):
        if self._rows == 0 and len(self._data) == 0:  # if non-initialized dataframe,adapts its shape based on the row
            self._columns = len(row)
            for i in range(self._columns):
                new_column = Column(f"Column{i + 1}")
                self._data.append(new_column)

        if len(row) != self._columns:
            raise ValueError("Invalid row length")
        for i in range(self._columns):
            self._data[i].data.append(row[i])
        self._rows += 1

    def add_column(self, column: Column):
        if len(column.data) != self._rows and self._rows != 0:
            raise ValueError("Invalid column length")
        self._data.append(column)
        self._columns += 1

    def remove_row(self, index: int | list[int]):
        # Removing multiple rows at once it was tricky of tricky since data shrinks in the iteration and indexes change
        if isinstance(index, int):
            if index >= self._rows:
                raise ValueError("Index out of range")
            for column in self._data:
                del column.data[index]
            self._rows -= 1
        elif isinstance(index, list):
            if len(index) > self._rows:
                raise ValueError("Too many indices for this dataframe")
            sorted_list = sorted(index)
            delta = 0
            for i in sorted_list:
                if i >= self._rows:
                    raise ValueError("Index out of range")
                for column in self._data:
                    del column.data[i - delta]
                delta += 1
                self._rows -= 1

    def shuffle(self, seed: int = 1):
        random.seed(seed)
        indices = list(range(self._rows))
        random.shuffle(indices)
        temp_data = DataFrame()

        # Initialize columns in temp_data with empty data
        for column in self._data:
            new_column = Column(column.name)
            new_column.type = column.type
            new_column.type = column.type
            new_column.data = []
            temp_data._data.append(new_column)
            temp_data._columns += 1

        # Add rows in shuffled order to temp_data
        for i in indices:
            for j in range(self._columns):
                temp_data._data[j].data.append(self._data[j].data[i])
            temp_data._rows += 1

        # Replace self with temp_data
        self._data = temp_data._data
        self._rows = temp_data._rows
        self._columns = temp_data._columns

    def count_occurrences(self, value_to_count):
        count = 0
        for column in self._data:
            count += column.data.count(value_to_count)
        return count

    def remove_column(self, name: str | int):
        if isinstance(name, int):
            name = self._data[name].name
        for column in self._data:
            if column.name == name:
                self._data.remove(column)
                self._columns -= 1
                return
        raise ValueError("Column not found")

    def get_column_names(self) -> list[str]:
        return [column.name for column in self._data]

    def set_columns_name(self, name_list: list[str]):
        if len(self._data) == 0 and len(name_list) > 0:  # for a non-initialized dataframe
            # Initialize columns with empty data
            for name in name_list:
                new_column = Column(name)
                self._data.append(new_column)
            self._columns = len(name_list)
        elif len(name_list) != self._columns:
            raise ValueError("Invalid name list length")

        for i, name in enumerate(name_list):    # Set the column names
            self._data[i].name = name

    def get_column(self, value: int | str) -> Column:
        if isinstance(value, int):
            if value >= self._columns:
                raise ValueError("Column index out of range")
            return self._data[value]
        elif isinstance(value, str):
            if value not in self.get_column_names():
                raise ValueError("Column not found")
            for column in self._data:
                if column.name == value:
                    return column

    def pop_column(self, value: int | str) -> list:
        """Returns the column as a list and removes it from the dataframe
        Args:
            value (int | str): The column name or index
        Returns:
            list: The column as a list
        Raises:
            ValueError: Column not found
            ValueError: Column index out of range
            """
        column_to_return = None
        if isinstance(value, int):
            if value >= self._columns:
                raise ValueError("Column index out of range")
            column_to_return = self._data[value]
        elif isinstance(value, str):
            if value not in self.get_column_names():
                raise ValueError(f"Column {value} not found")
            for column in self._data:
                if column.name == value:
                    column_to_return = column
                    break
        self.remove_column(value)
        return column_to_return

    def pop_row(self, index: int) -> list:
        """ Returns the row as a list and removes it from the dataframe"""
        if index >= self._rows:
            raise ValueError("Row index out of range")
        row_to_return = []
        for column in self._data:
            row_to_return.append(column.data[index])
        self.remove_row(index)
        return row_to_return

    def merge(self, other: 'DataFrame', horizontal: bool = True):
        """Merge another dataframe with this dataframe, either horizontally by rows or vertically by columns"""
        if horizontal:
            for i in range(other._rows):
                self.add_row(other[i])
        else:
            for i in range(other._columns):
                self.add_column(other.get_column(i))

    def to_csv(self, filename: str):
        with open(filename, 'w', encoding='utf-8', newline='\n') as f:
            # write the column headers
            header = [column.name for column in self._data]
            f.write(','.join(header) + '\n')
            # write the rows
            for i in range(self._rows):
                row = [str(column.data[i]) for column in self._data]
                f.write(','.join(row) + '\n')

    @classmethod
    def from_csv(cls, filename: str, sep: str = ",", header: bool = True):
        new_instance = cls()
        with open(filename, "r") as file:
            if header:
                headers = file.readline().strip().split(sep)
                new_instance._data = [Column(name) for name in headers]

            for line in file:
                row = line.strip().split(sep)
                for col_index, item in enumerate(row):
                    new_instance._data[col_index].data.append(item)

        new_instance._rows = len(new_instance._data[0].data)
        new_instance._columns = len(new_instance._data)
        return new_instance

    def __getitem__(self, indices):
        if isinstance(indices, int):
            # Singleton index is for accessing a row, return a Row proxy that can be indexed again for column access
            # in double indexing [][]. I really want to implement all the possible slicing operations, but this implies
            # modifying the underlying data structure for every operation in this and the proxy classes Column and Row
            if indices >= self._rows:
                raise IndexError("Row index out of range")
            return Row(self, indices)
        elif isinstance(indices, str):
            # Direct access to columns by name
            return self.column(indices)
        elif isinstance(indices, tuple):
            # Direct cell access via (row, col)
            row, col = indices
            if row >= self._rows or col >= self._columns:
                raise IndexError("Index out of range")
            return self._data[col].data[row]

    def __setitem__(self, key: int, value):
        """Only a Column object can be assigned by indexing obj[], this is bypassed by double-indexing [][] by
        overloaded__setitem__ in Column class"""
        if isinstance(key, int):
            if not isinstance(value, Column):
                raise ValueError("Value must be a Column object")
            print("Overloaded __setitem__")
            print(self._data[key])
            self._data[key] = value

    def __str__(self):
        return f"<DataFrame with {self._rows} rows and {self._columns} columns>"

    def __iter__(self) -> iter:
        # I have kinda simulated the iterrows() from pandas here, but overloaded the iterator instead
        cnt = 0
        for i in range(self._rows):
            cnt += 1
            yield cnt - 1, self[i]
